@@ -32,31 +32,31 @@ void QQmlTreeModelToTableModel::setModel(QAbstractItemModel *arg)
         const char *signal;
         const char *slot;
     };
-    static const Cx connections[] = {
+    const Cx connections[] = {
         { SIGNAL(destroyed(QObject*)),
           SLOT(modelHasBeenDestroyed()) },
         { SIGNAL(modelReset()),
           SLOT(modelHasBeenReset()) },
-        { SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&, const QVector<int>&)),
-          SLOT(modelDataChanged(const QModelIndex&, const QModelIndex&, const QVector<int>&)) },
+        { SIGNAL(dataChanged(QModelIndex,QModelIndex,QList<int>)),
+          SLOT(modelDataChanged(QModelIndex,QModelIndex,QList<int>)) },
 
-        { SIGNAL(layoutAboutToBeChanged(const QList<QPersistentModelIndex>&, QAbstractItemModel::LayoutChangeHint)),
-          SLOT(modelLayoutAboutToBeChanged(const QList<QPersistentModelIndex>&, QAbstractItemModel::LayoutChangeHint)) },
-        { SIGNAL(layoutChanged(const QList<QPersistentModelIndex>&, QAbstractItemModel::LayoutChangeHint)),
-          SLOT(modelLayoutChanged(const QList<QPersistentModelIndex>&, QAbstractItemModel::LayoutChangeHint)) },
+        { SIGNAL(layoutAboutToBeChanged(QList<QPersistentModelIndex>,QAbstractItemModel::LayoutChangeHint)),
+          SLOT(modelLayoutAboutToBeChanged(QList<QPersistentModelIndex>,QAbstractItemModel::LayoutChangeHint)) },
+        { SIGNAL(layoutChanged(QList<QPersistentModelIndex>,QAbstractItemModel::LayoutChangeHint)),
+          SLOT(modelLayoutChanged(QList<QPersistentModelIndex>,QAbstractItemModel::LayoutChangeHint)) },
 
-        { SIGNAL(rowsAboutToBeInserted(const QModelIndex&, int, int)),
-          SLOT(modelRowsAboutToBeInserted(const QModelIndex &, int, int)) },
-        { SIGNAL(rowsInserted(const QModelIndex&, int, int)),
-          SLOT(modelRowsInserted(const QModelIndex&, int, int)) },
-        { SIGNAL(rowsAboutToBeRemoved(const QModelIndex&, int, int)),
-          SLOT(modelRowsAboutToBeRemoved(const QModelIndex&, int, int)) },
-        { SIGNAL(rowsRemoved(const QModelIndex&, int, int)),
-          SLOT(modelRowsRemoved(const QModelIndex&, int, int)) },
-        { SIGNAL(rowsAboutToBeMoved(const QModelIndex&, int, int, const QModelIndex&, int)),
-          SLOT(modelRowsAboutToBeMoved(const QModelIndex&, int, int, const QModelIndex&, int)) },
-        { SIGNAL(rowsMoved(const QModelIndex&, int, int, const QModelIndex&, int)),
-          SLOT(modelRowsMoved(const QModelIndex&, int, int, const QModelIndex&, int)) },
+        { SIGNAL(rowsAboutToBeInserted(QModelIndex,int,int)),
+          SLOT(modelRowsAboutToBeInserted(QModelIndex,int,int)) },
+        { SIGNAL(rowsInserted(QModelIndex,int,int)),
+          SLOT(modelRowsInserted(QModelIndex,int,int)) },
+        { SIGNAL(rowsAboutToBeRemoved(QModelIndex,int,int)),
+          SLOT(modelRowsAboutToBeRemoved(QModelIndex,int,int)) },
+        { SIGNAL(rowsRemoved(QModelIndex,int,int)),
+          SLOT(modelRowsRemoved(QModelIndex,int,int)) },
+        { SIGNAL(rowsAboutToBeMoved(QModelIndex,int,int,QModelIndex,int)),
+          SLOT(modelRowsAboutToBeMoved(QModelIndex,int,int,QModelIndex,int)) },
+        { SIGNAL(rowsMoved(QModelIndex,int,int,QModelIndex,int)),
+          SLOT(modelRowsMoved(QModelIndex,int,int,QModelIndex,int)) },
         { nullptr, nullptr }
     };
 
@@ -69,6 +69,9 @@ void QQmlTreeModelToTableModel::setModel(QAbstractItemModel *arg)
         clearModelData();
         m_model = arg;
 
+        if (m_rootIndex.isValid() && m_rootIndex.model() != m_model)
+            m_rootIndex = QModelIndex();
+
         if (m_model) {
             for (const Cx *c = &connections[0]; c->signal; c++)
                 connect(m_model, c->signal, this, c->slot);
@@ -164,6 +167,11 @@ QVariant QQmlTreeModelToTableModel::headerData(int section, Qt::Orientation orie
     return m_model->headerData(section, orientation, role);
 }
 
+Qt::ItemFlags QQmlTreeModelToTableModel::flags(const QModelIndex &index) const
+{
+    return m_model->flags(mapToModel(index));
+}
+
 int QQmlTreeModelToTableModel::depthAtRow(int row) const
 {
     if (row < 0 || row >= m_items.size())
