@@ -1,30 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2020 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtQml module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL-EXCEPT$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2020 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 
 #include <QtTest/QtTest>
@@ -79,7 +54,7 @@ QDebug operator<<(QDebug dbg, const struct Event &event) {
     if (event.points.isEmpty())
         dbg << " @ " << event.mousePos << " global " << event.mousePosGlobal;
     else
-        dbg << ", " << event.points.count() << " touchpoints: " << event.points;
+        dbg << ", " << event.points.size() << " touchpoints: " << event.points;
     dbg << ')';
     return dbg;
 }
@@ -150,6 +125,12 @@ public:
         ++touchUngrabCount;
     }
 
+    void dumpEventList()
+    {
+        for (const auto &event : eventList)
+            qDebug() << event;
+    }
+
     bool event(QEvent *event) override {
         return QQuickItem::event(event);
     }
@@ -185,12 +166,22 @@ class GrabMonitor : public QObject
 {
 public:
     QObject *exclusiveGrabber = nullptr;
+    int transitionCount = 0;
     bool fromMouseEvent = false;
     bool canceled = false;
 
+    void reset()
+    {
+        exclusiveGrabber = nullptr;
+        transitionCount = 0;
+        fromMouseEvent = false;
+        canceled = false;
+    }
+
     void onGrabChanged(QObject *grabber, QPointingDevice::GrabTransition transition, const QPointerEvent *event, const QEventPoint &point)
     {
         qCDebug(lcTests) << grabber << transition << event << point << point.device();
+        ++transitionCount;
         switch (transition) {
         case QPointingDevice::GrabTransition::GrabExclusive:
             exclusiveGrabber = grabber;
@@ -247,6 +238,11 @@ private slots:
 
     void hoverEnabled();
     void implicitUngrab();
+    void touchCancelWillCancelMousePress();
+
+    void oneTouchInsideAndOneOutside();
+
+    void strayTouchDoesntAutograb();
 
 protected:
     bool eventFilter(QObject *, QEvent *event) override
@@ -717,7 +713,7 @@ void tst_TouchMouse::touchButtonOnFlickable()
 
     QTRY_COMPARE(eventItem2->touchUngrabCount, 1);
     qCDebug(lcTests) << "expected delivered events: press(touch) move(touch)" << eventItem2->eventList;
-    QCOMPARE(eventItem2->eventList.size(), 2);
+    QCOMPARE(eventItem2->eventList.size(), 3);
     QCOMPARE(eventItem2->eventList.at(1).type, QEvent::TouchUpdate);
     QCOMPARE(grabMonitor.exclusiveGrabber, flickable);
     // both EventItem and Flickable handled the actual touch, so synth-mouse doesn't happen
@@ -800,7 +796,7 @@ void tst_TouchMouse::buttonOnDelayedPressFlickable()
         qCDebug(lcTests) << "expected filtered events: actual TouchBegin and replayed TouchBegin" << filteredEventList;
         QTRY_COMPARE(eventItem1->eventList.size(), 1);
         QCOMPARE(eventItem1->eventList.at(0).type, QEvent::MouseButtonPress);
-        QCOMPARE(filteredEventList.count(), 2); // actual touch begin and replayed touch begin
+        QCOMPARE(filteredEventList.size(), 2); // actual touch begin and replayed touch begin
     }
 
     if (!releaseBeforeDelayIsOver) {
@@ -815,7 +811,7 @@ void tst_TouchMouse::buttonOnDelayedPressFlickable()
         if (scrollBeforeDelayIsOver) {
             QCOMPARE(eventItem1->eventList.size(), 0);
             qCDebug(lcTests) << "expected filtered events: 1 TouchBegin and 3 TouchUpdate" << filteredEventList;
-            QCOMPARE(filteredEventList.count(), 4);
+            QCOMPARE(filteredEventList.size(), 4);
         } else {
             qCDebug(lcTests) << "expected delivered events: press(mouse), move(mouse), move(mouse), ungrab(mouse)" << eventItem1->eventList;
             QCOMPARE(eventItem1->eventList.size(), 4);
@@ -823,7 +819,7 @@ void tst_TouchMouse::buttonOnDelayedPressFlickable()
             QCOMPARE(eventItem1->eventList.at(1).type, QEvent::MouseMove);
             QCOMPARE(eventItem1->eventList.last().type, QEvent::UngrabMouse);
             qCDebug(lcTests) << "expected filtered events: 2 TouchBegin and 3 TouchUpdate" << filteredEventList;
-            QCOMPARE(filteredEventList.count(), 5);
+            QCOMPARE(filteredEventList.size(), 5);
         }
 
         // flickable should have the touchpoint grab: it no longer relies on synth-mouse
@@ -844,17 +840,17 @@ void tst_TouchMouse::buttonOnDelayedPressFlickable()
         QCOMPARE(eventItem1->eventList.at(1).type, QEvent::MouseButtonRelease);
         QCOMPARE(eventItem1->eventList.last().type, QEvent::UngrabMouse);
         // QQuickWindow filters the delayed press and release
-        QCOMPARE(filteredEventList.count(), 4);
-        QCOMPARE(filteredEventList.at(filteredEventList.count() - 2).type, QEvent::TouchBegin);
+        QCOMPARE(filteredEventList.size(), 4);
+        QCOMPARE(filteredEventList.at(filteredEventList.size() - 2).type, QEvent::TouchBegin);
         QCOMPARE(filteredEventList.last().type, QEvent::TouchEnd);
     } else {
         // QQuickWindow filters the delayed press if there was one
         if (scrollBeforeDelayIsOver) {
             qCDebug(lcTests) << "expected filtered events: 1 TouchBegin, 3 TouchUpdate, 1 TouchEnd" << filteredEventList;
-            QCOMPARE(filteredEventList.count(), 5);
+            QCOMPARE(filteredEventList.size(), 5);
         } else {
             qCDebug(lcTests) << "expected filtered events: 2 TouchBegin, 3 TouchUpdate, 1 TouchEnd" << filteredEventList;
-            QCOMPARE(filteredEventList.count(), 6);
+            QCOMPARE(filteredEventList.size(), 6);
             QCOMPARE(filteredEventList.at(0).type, QEvent::TouchBegin);
             QCOMPARE(filteredEventList.last().type, QEvent::TouchEnd);
         }
@@ -969,7 +965,7 @@ void tst_TouchMouse::buttonOnTouch()
     touchSeq.press(0, p1, &window).press(1, p2, &window).commit();
     QQuickTouchUtils::flush(&window);
     QCOMPARE(button1->scale(), 1);
-    QCOMPARE(eventItem1->eventList.count(), 1);
+    QCOMPARE(eventItem1->eventList.size(), 1);
     QCOMPARE(eventItem1->eventList.at(0).type, QEvent::MouseButtonPress);
 
     p1 -= QPoint(10, 0);
@@ -1281,16 +1277,16 @@ void tst_TouchMouse::tapOnDismissiveTopMouseAreaClicksBottomOne()
     QTest::touchEvent(&window, device).release(0, p1, &window);
     QQuickTouchUtils::flush(&window);
 
-    QCOMPARE(bottomClickedSpy.count(), 1);
-    QCOMPARE(bottomDoubleClickedSpy.count(), 0);
+    QCOMPARE(bottomClickedSpy.size(), 1);
+    QCOMPARE(bottomDoubleClickedSpy.size(), 0);
 
     QTest::touchEvent(&window, device).press(0, p1, &window);
     QQuickTouchUtils::flush(&window);
     QTest::touchEvent(&window, device).release(0, p1, &window);
     QQuickTouchUtils::flush(&window);
 
-    QCOMPARE(bottomClickedSpy.count(), 1);
-    QCOMPARE(bottomDoubleClickedSpy.count(), 1);
+    QCOMPARE(bottomClickedSpy.size(), 1);
+    QCOMPARE(bottomDoubleClickedSpy.size(), 1);
 }
 
 /*
@@ -1444,6 +1440,7 @@ void tst_TouchMouse::hoverEnabled() // QTBUG-40856
     QQuickView window;
     QVERIFY(QQuickTest::showView(window, testFileUrl("hoverMouseAreas.qml")));
     QQuickItem *root = window.rootObject();
+    auto deliveryAgent = QQuickWindowPrivate::get(&window)->deliveryAgentPrivate();
 
     QQuickMouseArea *mouseArea1 = root->findChild<QQuickMouseArea*>("mouseArea1");
     QVERIFY(mouseArea1 != nullptr);
@@ -1453,11 +1450,11 @@ void tst_TouchMouse::hoverEnabled() // QTBUG-40856
 
     QSignalSpy enterSpy1(mouseArea1, SIGNAL(entered()));
     QSignalSpy exitSpy1(mouseArea1, SIGNAL(exited()));
-    QSignalSpy clickSpy1(mouseArea1, SIGNAL(clicked(QQuickMouseEvent *)));
+    QSignalSpy clickSpy1(mouseArea1, SIGNAL(clicked(QQuickMouseEvent*)));
 
     QSignalSpy enterSpy2(mouseArea2, SIGNAL(entered()));
     QSignalSpy exitSpy2(mouseArea2, SIGNAL(exited()));
-    QSignalSpy clickSpy2(mouseArea2, SIGNAL(clicked(QQuickMouseEvent *)));
+    QSignalSpy clickSpy2(mouseArea2, SIGNAL(clicked(QQuickMouseEvent*)));
 
     QPoint p1(150, 150);
     QPoint p2(150, 250);
@@ -1465,54 +1462,60 @@ void tst_TouchMouse::hoverEnabled() // QTBUG-40856
     // ------------------------- Mouse move to mouseArea1
     QTest::mouseMove(&window, p1);
 
-    QVERIFY(enterSpy1.count() == 1);
+    QVERIFY(enterSpy1.size() == 1);
     QVERIFY(mouseArea1->hovered());
     QVERIFY(!mouseArea2->hovered());
 
     // ------------------------- Touch click on mouseArea1
     QTest::touchEvent(&window, device).press(0, p1, &window);
+    deliveryAgent->flushFrameSynchronousEvents(&window);
 
-    QCOMPARE(enterSpy1.count(), 1);
-    QCOMPARE(enterSpy2.count(), 0);
-    QVERIFY(mouseArea1->pressed());
+    QCOMPARE(enterSpy1.size(), 1);
+    QCOMPARE(enterSpy2.size(), 0);
+    QVERIFY(mouseArea1->isPressed());
     QVERIFY(mouseArea1->hovered());
     QVERIFY(!mouseArea2->hovered());
 
     QTest::touchEvent(&window, device).release(0, p1, &window);
-    QVERIFY(clickSpy1.count() == 1);
+    deliveryAgent->flushFrameSynchronousEvents(&window);
+    QVERIFY(clickSpy1.size() == 1);
     QVERIFY(mouseArea1->hovered());
     QVERIFY(!mouseArea2->hovered());
 
     // ------------------------- Touch click on mouseArea2
     QTest::touchEvent(&window, device).press(0, p2, &window);
+    deliveryAgent->flushFrameSynchronousEvents(&window);
 
-    QVERIFY(mouseArea1->hovered());
+    QVERIFY(!mouseArea1->hovered());
     QVERIFY(mouseArea2->hovered());
-    QVERIFY(mouseArea2->pressed());
-    QCOMPARE(enterSpy1.count(), 1);
-    QCOMPARE(enterSpy2.count(), 1);
+    QVERIFY(mouseArea2->isPressed());
+    QCOMPARE(enterSpy1.size(), 1);
+    QCOMPARE(enterSpy2.size(), 1);
 
     QTest::touchEvent(&window, device).release(0, p2, &window);
+    deliveryAgent->flushFrameSynchronousEvents(&window);
 
-    QVERIFY(clickSpy2.count() == 1);
-    QVERIFY(mouseArea1->hovered());
+    QVERIFY(clickSpy2.size() == 1);
+    QVERIFY(!mouseArea1->hovered());
     QVERIFY(!mouseArea2->hovered());
-    QCOMPARE(exitSpy1.count(), 0);
-    QCOMPARE(exitSpy2.count(), 1);
+    QCOMPARE(exitSpy1.size(), 1);
+    QCOMPARE(exitSpy2.size(), 1);
 
     // ------------------------- Another touch click on mouseArea1
     QTest::touchEvent(&window, device).press(0, p1, &window);
+    deliveryAgent->flushFrameSynchronousEvents(&window);
 
-    QCOMPARE(enterSpy1.count(), 1);
-    QCOMPARE(enterSpy2.count(), 1);
-    QVERIFY(mouseArea1->pressed());
+    QCOMPARE(enterSpy1.size(), 2);
+    QCOMPARE(enterSpy2.size(), 1);
+    QVERIFY(mouseArea1->isPressed());
     QVERIFY(mouseArea1->hovered());
     QVERIFY(!mouseArea2->hovered());
 
     QTest::touchEvent(&window, device).release(0, p1, &window);
-    QCOMPARE(clickSpy1.count(), 2);
+    deliveryAgent->flushFrameSynchronousEvents(&window);
+    QCOMPARE(clickSpy1.size(), 2);
     QVERIFY(mouseArea1->hovered());
-    QVERIFY(!mouseArea1->pressed());
+    QVERIFY(!mouseArea1->isPressed());
     QVERIFY(!mouseArea2->hovered());
 }
 
@@ -1546,6 +1549,117 @@ void tst_TouchMouse::implicitUngrab()
     QCOMPARE(eventItem->eventList.at(0).type, QEvent::UngrabMouse);
     QTest::touchEvent(&window, device).release(0, p1);   // clean up potential state
 }
+
+void tst_TouchMouse::touchCancelWillCancelMousePress()
+{
+    QQuickView window;
+    QVERIFY(QQuickTest::showView(window, testFileUrl("singleitem.qml")));
+    QQuickItem *root = window.rootObject();
+    QVERIFY(root != nullptr);
+
+    EventItem *eventItem = root->findChild<EventItem*>("eventItem1");
+    eventItem->acceptMouse = true;
+    eventItem->setAcceptTouchEvents(false);
+    QPoint p1(20, 20);
+
+    // Begin a new touch, that gets converted to a mouse press
+    QTest::touchEvent(&window, device).press(0, p1);
+    QCOMPARE(eventItem->eventList.size(), 1);
+    QCOMPARE(eventItem->eventList.at(0).type, QEvent::MouseButtonPress);
+
+    // Cancel it...
+    QTouchEvent cancelEvent(QEvent::TouchCancel, device);
+    QCoreApplication::sendEvent(&window, &cancelEvent);
+    QCOMPARE(eventItem->eventList.size(), 3);
+    QCOMPARE(eventItem->eventList.at(1).type, QEvent::TouchCancel);
+    QCOMPARE(eventItem->eventList.at(2).type, QEvent::UngrabMouse);
+
+    // Begin a second touch. Since the last one was cancelled, this
+    // should end up as a new mouse press on the target item.
+    QTest::touchEvent(&window, device).press(0, p1);
+    QVERIFY(eventItem->eventList.size() >= 5);
+    QCOMPARE(eventItem->eventList.at(3).type, QEvent::MouseButtonPress);
+
+    QTest::touchEvent(&window, device).release(0, p1);   // clean up potential state
+}
+
+void tst_TouchMouse::oneTouchInsideAndOneOutside() // QTBUG-102996
+{
+    QQuickView window;
+    QVERIFY(QQuickTest::showView(window, testFileUrl("oneMouseArea.qml")));
+    QQuickItem *root = window.rootObject();
+    QVERIFY(root);
+    QQuickMouseArea *ma = root->findChild<QQuickMouseArea*>();
+    QVERIFY(ma);
+
+    // Press the MouseArea
+    QPoint p1 = ma->mapToScene(ma->boundingRect().center()).toPoint();
+    QTest::touchEvent(&window, device).press(1, p1);
+    QQuickTouchUtils::flush(&window);
+    QVERIFY(ma->isPressed());
+
+    // Tap outside the MouseArea with a second finger
+    QPoint p2(100, 100);
+    QTest::touchEvent(&window, device).stationary(1).press(2, p2);
+    QQuickTouchUtils::flush(&window);
+    QTest::touchEvent(&window, device).stationary(1).release(2, p2);
+    QQuickTouchUtils::flush(&window);
+    QVERIFY(ma->isPressed());
+
+    // Press again outside the MouseArea with a second finger
+    QTest::touchEvent(&window, device).stationary(1).press(2, p2);
+
+    // Release the first finger: MouseArea should be released
+    QTest::touchEvent(&window, device).release(1, p1).stationary(2);
+    QQuickTouchUtils::flush(&window);
+    QCOMPARE(ma->isPressed(), false);
+
+    // Release the second finger
+    QTest::touchEvent(&window, device).release(2, p2);
+    QQuickTouchUtils::flush(&window);
+}
+
+void tst_TouchMouse::strayTouchDoesntAutograb() // QTBUG-107867
+{
+    QQuickView window;
+    QVERIFY(QQuickTest::showView(window, testFileUrl("singleitem.qml")));
+    QQuickItem *root = window.rootObject();
+    QVERIFY(root);
+    EventItem *eventItem = root->findChild<EventItem*>();
+    QVERIFY(eventItem);
+    // This item accepts (synth-)mouse events but NOT touch
+    eventItem->acceptMouse = true;
+    QCOMPARE(eventItem->acceptTouchEvents(), false); // the default in Qt 6
+    QPoint p1(6, 6);
+    grabMonitor.reset();
+
+    // Begin a new touch, that gets converted to a mouse press
+    QTest::touchEvent(&window, device).press(0, p1);
+    QQuickTouchUtils::flush(&window);
+    qCDebug(lcTests) << "after touch press:" << eventItem->eventList;
+    QCOMPARE(eventItem->eventList.size(), 1);
+    QCOMPARE(eventItem->eventList.at(0).type, QEvent::MouseButtonPress);
+    QCOMPARE(grabMonitor.exclusiveGrabber, eventItem);
+
+    // Drag
+    for (int i = 0; i < 3; ++i) {
+        QTest::touchEvent(&window, device).move(0, p1 + QPoint(i * 5, i * 5), &window);
+        QQuickTouchUtils::flush(&window);
+        QCOMPARE(grabMonitor.transitionCount, 1); // no new grab
+        QCOMPARE(eventItem->eventList.size(), i + 2);
+        QCOMPARE(eventItem->eventList.last().type, QEvent::MouseMove);
+    }
+
+    // Press an extra point: EventItem should see nothing
+    QTest::touchEvent(&window, device).stationary(0).press(1, p1);
+    QQuickTouchUtils::flush(&window);
+    qCDebug(lcTests) << "after press of second touchpoint:" << eventItem->eventList;
+    QCOMPARE(eventItem->eventList.size(), 4);
+    QCOMPARE(grabMonitor.transitionCount, 1); // no new grab
+
+    QTest::touchEvent(&window, device).release(0, p1).release(1, p1);
+}
+
 QTEST_MAIN(tst_TouchMouse)
 
 #include "tst_touchmouse.moc"
